Require Import Functors.
Require Import List.
Require Import FunctionalExtensionality.
Require Import MonadLib.

Section Names.

  (* ============================================== *)
  (* ENVIRONMENTS                                   *)
  (* ============================================== *)

  Definition Env (A : Set) := list A.

  Fixpoint lookup {A: Set} (As : Env A) (i : nat) : option A :=
    match As, i with
      | nil, _  => None
      | cons _ As', S i' => lookup As' i'
      | cons a _, 0 => Some a
    end.

  Fixpoint insert (A : Set) (e : A) (env : Env A) : Env A :=
    match env in list _ return Env A with
      | nil => cons e (nil)
      | cons e' env' => cons e' (@insert _ e env')
    end.

  Definition empty A : Env A := nil.

  Lemma length_insert {A : Set} :
    forall (a : A) (As : Env A), length (insert _ a As) = S (length As).
  Proof.
    induction As; simpl; auto.
  Qed.

  Lemma lookup_insert {A : Set} :
    forall (a : A) (As : Env A),
      lookup (insert _ a As) (length As) = Some a.
  Proof.
    induction As; simpl; auto.
  Qed.

  Lemma lookup_insert_lt {A : Set} :
    forall (a : A) (As : Env A) (i : nat),
      i < (length As) -> lookup (insert _ a As) i = lookup As i.
  Proof.
    induction As; destruct i; simpl; intros; auto with arith.
    inversion H.
  Qed.

  Lemma lookup_Some_lt {A : Set} :
    forall (a : A) (i : nat) (As : Env A),
      lookup As i = Some a -> i < (length As).
  Proof.
    induction i; destruct As; simpl; intros; try discriminate;
      auto with arith.
  Qed.

  Lemma lookup_Some_insert {A : Set} :
    forall (a : A) (i : nat) (As : Env A),
      lookup As i = Some a ->
      forall (a' : A), lookup (insert _ a' As) i = lookup As i.
  Proof.
    intros; apply lookup_insert_lt.
    eapply lookup_Some_lt; eauto.
  Qed.

  (* ============================================== *)
  (* TYPES                                          *)
  (* ============================================== *)


  (** SuperFunctor for Types. **)
  Variable DT : Set -> Set.
  Context {Fun_DT : Functor DT}.
  Definition DType := UP'_F DT.


  (* ============================================== *)
  (* VALUES                                         *)
  (* ============================================== *)


  (** SuperFunctor for Values. **)

  Variable V : Set -> Set.
  Context {Fun_V : Functor V}.
  Definition Value := UP'_F V.

  (** ERROR VALUES **)

   Inductive StuckValue (A : Set) : Set :=
    | Stuck : nat -> StuckValue A.

   Context {Sub_StuckValue_V : StuckValue :<: V}.

   Definition Stuck_fmap : forall (A B : Set) (f : A -> B),
     StuckValue A -> StuckValue B := fun A B _ e =>
       match e with
         | Stuck n => Stuck _ n
       end.

   Global Instance Stuck_Functor : Functor StuckValue :=
     {| fmap := Stuck_fmap |}.
     destruct a; reflexivity.
     (* fmap_id *)
     destruct a; reflexivity.
   Defined.

   (* Constructor + Universal Property. *)
   Context {WF_SubStuckValue_V : WF_Functor _ _ Sub_StuckValue_V _ _}.

   Definition stuck' (n : nat) : Value := inject' (Stuck _ n).
   Definition stuck (n : nat) : Fix V := proj1_sig (stuck' n).

   Global Instance UP'_stuck {n : nat} :
     Universal_Property'_fold (stuck n) := proj2_sig (stuck' n).

   (* Induction Principle for Stuckor Values. *)

  Definition ind_alg_Stuck (P : Fix V -> Prop)
    (H : forall n, P (stuck n))
      (e : StuckValue (sig P)) : sig P :=
    match e with
      | Stuck n => exist P (stuck n) (H n)
    end.

  Definition ind_palg_Stuck (Name : Set) (P : Fix V -> Prop) (H : forall n, P (stuck n)) :
    PAlgebra Name StuckValue V P.
  Proof.
    econstructor 1 with (p_algebra := ind_alg_Stuck P H).
    destruct e; simpl; unfold stuck, stuck'; simpl;
      rewrite wf_functor; reflexivity.
  Qed.

   (** BOTTOM VALUES **)

   Inductive BotValue (A : Set) : Set :=
   | Bot : BotValue A.

   Context {Sub_BotValue_V : BotValue :<: V}.

   Definition Bot_fmap : forall (A B : Set) (f : A -> B),
     BotValue A -> BotValue B := fun A B _ _ => Bot _.

   Global Instance Bot_Functor : Functor BotValue :=
     {| fmap := Bot_fmap |}.
     destruct a; reflexivity.
     (* fmap_id *)
     destruct a. reflexivity.
   Defined.

   (* Constructor + Universal Property. *)
   Context {WF_SubBotValue_V : WF_Functor _ _ Sub_BotValue_V _ _}.

   Definition bot' : Value := inject' (Bot _).
   Definition bot : Fix V := proj1_sig bot'.
   Global Instance UP'_bot : Universal_Property'_fold bot :=
     proj2_sig bot'.

  Definition ind_alg_Bot (P : Fix V -> Prop)
    (H : P bot)
      (e : BotValue (sig P)) : sig P :=
    match e with
      | Bot => exist P bot H
    end.

  (* Constructor Testing for Bottom Values. *)

  Definition isBot : Fix V -> bool :=
    fun exp =>
      match project exp with
       | Some Bot  => true
       | None      => false
      end.

  Definition ind_palg_Bot (Name : Set) (P : Fix V -> Prop) (H : P bot) :
    PAlgebra Name BotValue V P.
  Proof.
    econstructor 1 with (p_algebra := ind_alg_Bot P H).
    destruct e; simpl; unfold bot, bot'; simpl;
      rewrite wf_functor; reflexivity.
  Qed.

  (* ============================================== *)
  (* EXPRESSIONS                                    *)
  (* ============================================== *)

  (** SuperFunctor for Expressions. **)
  Variable E : Set -> Set.
  Context {Fun_E : Functor E}.
  Definition Exp := UP'_F E.

  (* ============================================== *)
  (* OPERATIONS                                     *)
  (* ============================================== *)


  (** TYPING **)

  Variable MT : Set -> Set.

  Definition typeofR  := MT DType.
  Inductive TypeofName : Set := typeofname.
  Context {Typeof_E : forall T,
    FAlgebra TypeofName T typeofR E}.
  Definition typeof :=
    mfold _ (fun _ => @f_algebra _ _ _ _ (Typeof_E _)).

  (** EVALUATION **)

  Definition evalR : Set := Env Value -> Value.

  Inductive EvalName := evalname.

  Context {eval_E : forall T, FAlgebra EvalName T evalR E}.
  Definition eval := mfold _ (fun _ => @f_algebra _ _ _ _ (eval_E _)).

  Context {beval_E : FAlgebra EvalName Exp evalR E}.

  Definition beval (n: nat) :=
    boundedFix_UP n (@f_algebra _ _ _ _ beval_E) (fun _ => bot').

  Variable ME : Set -> Set.
  Context {Functor_ME : Functor ME}.
  Context {Monad_ME : Monad ME}.
  Context {Fail_ME : FailMonad ME}.

  Definition evalMR := ME Value.

  Context {evalM_E : forall T, FAlgebra EvalName T evalMR E}.
  Definition evalM := mfold _ (fun _ => @f_algebra _ _ _ _ (evalM_E _)).

  Context {bevalM_E : FAlgebra EvalName Exp evalMR E}.

  Definition bevalM (n: nat) :=
    boundedFix_UP n (@f_algebra _ _ _ _ bevalM_E) fail.

  (** DTYPE EQUALITY **)

  Definition eq_DTypeR := DType -> bool.
  Inductive eq_DTypeName := eq_dtypename.
  Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T eq_DTypeR DT}.
  Definition eq_DType := mfold _ (fun _ => @f_algebra _ _ _ _ (eq_DType_DT _)).

  Definition eq_DType_eq_P (d : Fix DT) (d_UP' : Universal_Property'_fold d) := forall d2,
    eq_DType d d2 = true -> d = proj1_sig d2.
  Inductive eq_DType_eqName := eq_dtype_eqname.
  Context {eq_DType_eq_DT : PAlgebra eq_DType_eqName DT DT (UP'_P eq_DType_eq_P)}.
  Lemma eq_DType_eq : forall (d1 : DType), eq_DType_eq_P (proj1_sig d1) (proj2_sig d1).
    intro; eapply (proj2_sig (Ind (P := UP'_P eq_DType_eq_P) (proj1_sig d1) (proj2_sig d1))).
  Qed.

  Definition eq_DType_neq_P (d : Fix DT) (d_UP' : Universal_Property'_fold d) := forall d2,
    eq_DType d d2 = false -> d <> proj1_sig d2.
  Inductive eq_DType_neqName := eq_dtype_neqname.
  Context {eq_DType_neq_D : PAlgebra eq_DType_neqName DT DT (UP'_P eq_DType_neq_P)}.
  Lemma eq_DType_neq : forall (d1 : DType), eq_DType_neq_P (proj1_sig d1) (proj2_sig d1).
    intro; eapply (proj2_sig (Ind (P := UP'_P eq_DType_neq_P) (proj1_sig d1) (proj2_sig d1))).
  Qed.

  (** PRETTY PRINTING **)

  Require Import String.
  Require Import Ascii.

  Definition DTypePrintR := string.
  Inductive DTypePrintName := dtypeprintname.
  Context {DTypePrint_DT : forall T, FAlgebra DTypePrintName T DTypePrintR DT}.
  Definition DTypePrint := mfold _ (fun _ => @f_algebra _ _ _ _ (DTypePrint_DT _)).

  Definition ExpPrintR := nat -> string.
  Inductive ExpPrintName := expprintname.
  Context {ExpPrint_E : forall T, FAlgebra ExpPrintName T ExpPrintR E}.
  Definition ExpPrint e := mfold _ (fun _ => @f_algebra _ _ _ _ (ExpPrint_E _)) e 48.

  Definition ValuePrintR := string.
  Inductive ValuePrintName := valueprintname.
  Context {ValuePrint_V : forall T, FAlgebra ValuePrintName T ValuePrintR V}.
  Definition ValuePrint := mfold _ (fun _ => @f_algebra _ _ _ _ (ValuePrint_V _)).

  (* Printers for Bot and Stuck *)
   Global Instance MAlgebra_ValuePrint_BotValue T : FAlgebra ValuePrintName T ValuePrintR BotValue :=
     {| f_algebra :=  fun _ _ => append "bot" "" |}.

   Global Instance MAlgebra_ValuePrint_StuckValue T : FAlgebra ValuePrintName T ValuePrintR StuckValue :=
     {| f_algebra :=
       fun _ e => match e with
                      | Stuck n => append "Stuck " (String (ascii_of_nat (n + 48)) EmptyString)
                    end|}.

  (* ============================================== *)
  (* PREDICATE LIFTERS FOR LISTS                    *)
  (* ============================================== *)

    (* Unary Predicates.*)
    Inductive P_Env {A : Set} (P : A -> Prop) : forall (env : Env A), Prop :=
    | P_Nil : P_Env P nil
    | P_Cons : forall a (As : Env _), P a -> P_Env P As ->
      P_Env P (cons a As).

    Lemma P_Env_lookup : forall A (env : Env A) P,
      P_Env P env ->
      forall n v,
        lookup env n = Some v -> P v.
      intros A env P P_env; induction P_env;
        destruct n; simpl; intros; try discriminate.
      injection H0; intros; subst; eauto.
      eauto.
    Qed.

    Lemma P_Env_Lookup : forall A (env : Env A) (P : A -> Prop),
      (forall n v,
        lookup env n = Some v -> P v) ->
      P_Env P env.
      intros A env P H; induction env; constructor.
      eapply (H 0); eauto.
      apply IHenv; intros; eapply (H (S n)); eauto.
    Qed.

    Lemma P_Env_insert : forall A (env : Env A) (P : A -> Prop),
      P_Env P env -> forall a, P a -> P_Env P (insert _ a env).
      induction env; simpl; intros; constructor; eauto.
      inversion H; subst; eauto.
      eapply IHenv; inversion H; eauto.
    Qed.

    (* Binary Predicates.*)
    Inductive P2_Env {A B : Set} (P : A -> B -> Prop) : forall (env : Env A) (env : Env B), Prop :=
    | P2_Nil : P2_Env P nil nil
    | P2_Cons : forall a b (As : Env _) (Bs : Env _), P a b -> P2_Env P As Bs ->
      P2_Env P (cons a As) (cons b Bs).

    Lemma P2_Env_lookup : forall A B (env : Env A) (env' : Env B) P,
      P2_Env P env env' ->
      forall n v,
        lookup env n = Some v -> exists v', lookup env' n = Some v' /\
          P v v'.
      intros A B env env' P P_env_env'; induction P_env_env';
        destruct n; simpl; intros; try discriminate.
      exists b; injection H0; intros; subst; split; eauto.
      eauto.
    Qed.

    Lemma P2_Env_lookup' : forall A B (env : Env A) (env' : Env B) P,
      P2_Env P env env' ->
      forall n v,
        lookup env' n = Some v -> exists v', lookup env n = Some v' /\
          P v' v.
      intros A B env env' P P_env_env'; induction P_env_env';
        destruct n; simpl; intros; try discriminate.
      eexists; injection H0; intros; subst; split; eauto.
      eauto.
    Qed.

    Lemma P2_Env_Nlookup : forall A B (env : Env A) (env' : Env B) P,
      P2_Env P env env' ->
      forall n,
        lookup env n = None -> lookup env' n = None.
      intros A B env env' P P_env_env'; induction P_env_env';
        destruct n; simpl; intros; try discriminate; auto.
    Qed.

    Lemma P2_Env_Nlookup' : forall A B (env : Env A) (env' : Env B) P,
      P2_Env P env env' ->
      forall n,
        lookup env' n = None -> lookup env n = None.
      intros A B env env' P P_env_env'; induction P_env_env';
        destruct n; simpl; intros; try discriminate; auto.
    Qed.

    Lemma P2_Env_length : forall A B (env : Env A) (env' : Env B) P,
      P2_Env P env env' -> List.length env = List.length env'.
      intros; induction H; simpl; congruence.
    Qed.

    Lemma P2_Env_insert : forall A B (env : Env A) (env' : Env B) (P : A -> B -> Prop),
      P2_Env P env env' ->
      forall a b, P a b -> P2_Env P (insert _ a env) (insert _ b env').  
      intros; induction H; simpl; constructor; eauto.
      constructor.
    Qed.

    Fixpoint replace_el {A : Set} (n : nat) (a : A) (l : list A) :=
      match n, l with
        | 0, cons _ l' => cons a l'
        | S n', cons a' l' => cons a' (replace_el n' a l')
        | _, _ => nil
      end.

    Lemma P2_Env_replace_el : forall (A B : Set) (P : A -> B -> Prop) (n : nat)
      (env : Env A) (env' : Env B),
      P2_Env P env env' ->
      forall a b, P a b -> (lookup env' n = Some b) ->
        P2_Env P (replace_el n a env) env'.
    Proof.
      induction n; intros; destruct H; simpl in *; try discriminate.
      injection H1; intros; subst; constructor; auto.
      constructor; eauto.
    Qed.

    (* Need this better induction principle when we're reasoning about
       Functors that use P2_Envs. *)
    Definition P2_Env_ind' :=
    fun (A B : Set) (P : A -> B -> Prop) (P0 : forall As Bs, P2_Env P As Bs -> Prop)
      (f : P0 _ _ (P2_Nil _))
      (f0 : forall (a : A) (b : B) (As : Env A) (Bs : Env B) (ABs : P2_Env P As Bs)
        (P_a_b : P a b), P0 _ _ ABs -> P0 _ _ (P2_Cons P a b As Bs P_a_b ABs)) =>
        fix F (env : Env A) (env0 : Env B) (p : P2_Env P env env0) {struct p} :
        P0 env env0 p :=
        match p in (P2_Env _ env1 env2) return (P0 env1 env2 p) with
        | P2_Nil => f
        | P2_Cons a b As Bs y p0 => f0 a b As Bs p0 y (F As Bs p0)
      end.

    Lemma P2_Env_P_P' : forall A B (env : Env A) (env' : Env B) (P P' : A -> B -> Prop),
      P2_Env P env env' ->
      (forall a b, P a b -> P' a b) ->
      P2_Env P' env env'.
    Proof.
      intros until P'; intro P_env_env'.
      induction P_env_env'; constructor; eauto.
    Qed.

  (* ============================================== *)
  (* SUBVALUE RELATION                              *)
  (* ============================================== *)

    Record SubValue_i : Set := mk_SubValue_i
      {sv_a : Value;
        sv_b : Value}.

    (** SuperFunctor for SubValue Relation. **)

    Variable SV : (SubValue_i -> Prop) -> SubValue_i -> Prop.
    Definition SubValue := iFix SV.
    Definition SubValueC V1 V2:= SubValue (mk_SubValue_i V1 V2).
    Variable funSV : iFunctor SV.

    (** Subvalue is reflexive **)
    Inductive SubValue_refl (A : SubValue_i -> Prop) : SubValue_i -> Prop :=
      SV_refl : forall v v',
        proj1_sig v = proj1_sig v' ->
        SubValue_refl A (mk_SubValue_i v v').

    Definition ind_alg_SV_refl (P : SubValue_i -> Prop)
      (H : forall v v', proj1_sig v = proj1_sig v' -> P (mk_SubValue_i v v'))
      i (e : SubValue_refl P i) : P i :=
      match e in SubValue_refl _ i return P i with
        | SV_refl v v' eq_v => H v v' eq_v
      end.

    Definition SV_refl_ifmap (A B : SubValue_i -> Prop) i (f : forall i, A i -> B i)
      (SV_a : SubValue_refl A i) : SubValue_refl B i :=
      match SV_a in (SubValue_refl _ s) return (SubValue_refl B s)
        with
        | SV_refl v v' H => SV_refl B v v' H
      end.

    Global Instance iFun_SV_refl : iFunctor SubValue_refl.
      constructor 1 with (ifmap := SV_refl_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable Sub_SV_refl_SV : Sub_iFunctor SubValue_refl SV.

    (** Bot is Bottom element for this relation **)
    Inductive SubValue_Bot (A : SubValue_i -> Prop) : SubValue_i -> Prop :=
      SV_Bot : forall v v',
        proj1_sig v = inject (Bot _) ->
        SubValue_Bot A (mk_SubValue_i v v').

    Definition ind_alg_SV_Bot (P : SubValue_i -> Prop)
      (H : forall v v' v_eq, P (mk_SubValue_i v v'))
      i (e : SubValue_Bot P i) : P i :=
      match e in SubValue_Bot _ i return P i with
        | SV_Bot v v' v_eq => H v v' v_eq
      end.

    Definition SV_Bot_ifmap (A B : SubValue_i -> Prop) i (f : forall i, A i -> B i)
      (SV_a : SubValue_Bot A i) : SubValue_Bot B i :=
      match SV_a in (SubValue_Bot _ s) return (SubValue_Bot B s)
        with
        | SV_Bot v v' H => SV_Bot B v v' H
      end.

    Global Instance iFun_SV_Bot : iFunctor SubValue_Bot.
      constructor 1 with (ifmap := SV_Bot_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable Sub_SV_Bot_SV : Sub_iFunctor SubValue_Bot SV.

    (* Inversion principle for Bottom SubValues. *)
    Definition SV_invertBot_P (i : SubValue_i) :=
      proj1_sig (sv_b i) = bot -> proj1_sig (sv_a i) = bot.

    Inductive SV_invertBot_Name := ece_invertbot_name.
    Context {SV_invertBot_SV :
      iPAlgebra SV_invertBot_Name SV_invertBot_P SV}.

    Global Instance SV_invertBot_refl :
      iPAlgebra SV_invertBot_Name SV_invertBot_P (SubValue_refl).
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertBot_P.
      inversion H; subst; simpl; congruence.
    Defined.

    Global Instance SV_invertBot_Bot :
      iPAlgebra SV_invertBot_Name SV_invertBot_P SubValue_Bot.
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertBot_P.
      inversion H; subst; simpl; eauto.
    Defined.

    Definition SV_invertBot := ifold_ SV _ (ip_algebra (iPAlgebra := SV_invertBot_SV)).
    (* End Inversion principle for SubValue.*)

    (* Projection doesn't affect SubValue Relation.*)

     Definition SV_proj1_b_P (i :SubValue_i) :=
      forall b' H, b' = proj1_sig (sv_b i) ->
      SubValueC (sv_a i) (exist _ b' H).

     Inductive SV_proj1_b_Name := sv_proj1_b_name.
     Context {SV_proj1_b_SV :
       iPAlgebra SV_proj1_b_Name SV_proj1_b_P SV}.

     Definition SV_proj1_b :=
       ifold_ SV _ (ip_algebra (iPAlgebra := SV_proj1_b_SV)).

     Global Instance SV_proj1_b_refl :
       iPAlgebra SV_proj1_b_Name SV_proj1_b_P SubValue_refl.
       econstructor; intros.
       unfold iAlgebra; intros; unfold SV_proj1_b_P.
       inversion H; subst; simpl; intros.
       apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; simpl; congruence.
     Defined.

    Global Instance SV_proj1_b_Bot :
      iPAlgebra SV_proj1_b_Name SV_proj1_b_P SubValue_Bot.
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_proj1_b_P.
      inversion H; subst; simpl; eauto.
      intros; revert H1; rewrite H2; intros.
      apply inject_i.
      constructor; assumption.
    Defined.

    Definition SV_proj1_a_P (i : SubValue_i) :=
      forall a' H, proj1_sig (sv_a i) = a' ->
        SubValueC (exist _ a' H) (sv_b i).

    Inductive SV_proj1_a_Name := sv_proj1_a_name.
    Context {SV_proj1_a_SV :
      iPAlgebra SV_proj1_a_Name SV_proj1_a_P SV}.

    Definition SV_proj1_a :=
      ifold_ SV _ (ip_algebra (iPAlgebra := SV_proj1_a_SV)).

    Global Instance SV_proj1_a_refl :
      iPAlgebra SV_proj1_a_Name SV_proj1_a_P SubValue_refl.
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_proj1_a_P.
      inversion H; subst; simpl; intros.
      revert H1; rewrite <- H2; intros.
      apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; simpl; eauto.
    Defined.

    Global Instance SV_proj1_a_Bot :
      iPAlgebra SV_proj1_a_Name SV_proj1_a_P SubValue_Bot.
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_proj1_a_P.
      inversion H; subst; simpl; eauto.
      intros; revert H1; rewrite <- H2, H0; intros.
      apply inject_i.
      constructor; reflexivity.
    Defined.

    (** SubValue lifted to Environments **)
    Definition Sub_Environment (env env' : Env _) :=
      P2_Env SubValueC env env'.

    Lemma Sub_Environment_refl : forall (env : Env _),
      Sub_Environment env env.
      induction env; econstructor; eauto.
      apply (inject_i (subGF := Sub_SV_refl_SV));
        constructor; reflexivity.
    Qed.

  (* ============================================== *)
  (* EVALUATION IS CONTINUOUS                        *)
  (* ============================================== *)

    (** Helper property for proof of continuity of evaluation. **)
    Definition eval_continuous_Exp_P (e : Fix E)
      (e_UP' : Universal_Property'_fold e) :=
      forall (m : nat),
        (forall (e0 : Exp)
          (gamma gamma' : Env Value) (n : nat),
          Sub_Environment gamma gamma' ->
          m <= n -> SubValueC (beval m e0 gamma) (beval n e0 gamma')) ->
        forall (gamma gamma' : Env Value) (n : nat),
          Sub_Environment gamma gamma' ->
          m <= n ->
          SubValueC (beval (S m) (exist _ _ e_UP') gamma) (beval (S n) (exist _ _ e_UP') gamma').

    Inductive EC_ExpName := ec_expname.

    Variable eval_continuous_Exp_E : PAlgebra EC_ExpName E E (UP'_P eval_continuous_Exp_P).

    (** Bounded Evaluation is continuous. **)
    Lemma beval_continuous : forall m,
      forall (e : Exp) (gamma gamma' : Env _),
        forall n (Sub_G_G' : Sub_Environment gamma gamma'),
          m <= n ->
          SubValueC (beval m e gamma) (beval n e gamma').
      induction m; simpl.
      intros; eapply in_ti; eapply inj_i; econstructor; simpl; eauto.
      intros; destruct n; try (inversion H; fail).
      assert (m <= n) as le_m_n0 by auto with arith; clear H.
      revert m IHm gamma gamma' n Sub_G_G' le_m_n0.
      fold (eval_continuous_Exp_P (proj1_sig e)).
      apply (proj2_sig (Ind (P := UP'_P eval_continuous_Exp_P) _ (proj2_sig e))).
    Qed.

  (* ============================================== *)
  (* MONADIC EVALUATION IS CONTINUOUS               *)
  (* ============================================== *)

    Record SubValueM_i : Set := mk_SubValueM_i
      {svm_a : ME Value;
       svm_b : ME Value}.

    (** SuperFunctor for SubValue Relation. **)

    Variable SVM : (SubValueM_i -> Prop) -> SubValueM_i -> Prop.
    Definition SubValueM := iFix SVM.
    Definition SubValueMC V1 V2 := SubValueM (mk_SubValueM_i V1 V2).
    Variable funSVM : iFunctor SVM.

    (** Subvalue for Basic Monads. **)

    Inductive SubValueM_base (A : SubValueM_i -> Prop) : SubValueM_i -> Prop :=
    | SVM_Fail : forall me, SubValueM_base A (mk_SubValueM_i fail me)
    | SVM_Return : forall v, SubValueM_base A (mk_SubValueM_i (return_ v) (return_ v)).
    (* | SVM_Bind : forall me me' k k',
      A (mk_SubValueM_i me me') ->
      (forall (i : Value), A (mk_SubValueM_i (k i) (k' i))) ->
      SubValueM_base A (mk_SubValueM_i (me >>= k) (me' >>= k')). *)

    Definition ind_alg_SVM_base (P : SubValueM_i -> Prop)
      (H_Fail : forall me, P (mk_SubValueM_i fail me))
      (H_Return : forall v, P (mk_SubValueM_i (return_ v) (return_ v)))
      (*(H_Bind : forall me me' k k',
        P (mk_SubValueM_i me me') ->
        (forall (i : Value), P (mk_SubValueM_i (k i) (k' i))) ->
        P (mk_SubValueM_i (me >>= k) (me' >>= k'))) *)
      (i : SubValueM_i)
      (e : SubValueM_base P i)
      :
      P i :=
      match e in SubValueM_base _ i return P i with
        | SVM_Fail me => H_Fail me
        | SVM_Return v => H_Return v
        (*| SVM_Bind me me' k k' P_me_me' P_k_k' =>
          H_Bind me me' k k' P_me_me' P_k_k' *)
      end.

    Definition SVM_base_ifmap
      (A B : SubValueM_i -> Prop)
      (i : SubValueM_i)
      (f : forall i, A i -> B i)
      (SVM_a : SubValueM_base A i) : SubValueM_base B i :=
      match SVM_a in SubValueM_base _ i return SubValueM_base B i with
        | SVM_Fail me => SVM_Fail B me
        | SVM_Return me => SVM_Return B me
        (* | SVM_Bind me me' k k' A_me_me' A_k_k' =>
          SVM_Bind B me me' k k' (f _ A_me_me')
            (fun i => f _ (A_k_k' i)) *)
      end.

    Global Instance iFun_SVM_base : iFunctor SubValueM_base.
      constructor 1 with (ifmap := SVM_base_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; auto.
    Defined.

    Context {Sub_SVM_base_SVM : Sub_iFunctor SubValueM_base SVM}.

    (** Helper property for proof of continuity of evaluation. **)
    Definition bevalM_continuous_Exp_P (e : Fix E)
      (e_UP' : Universal_Property'_fold e) :=
      forall (m : nat),
        (forall (e0 e' : Exp) (n : nat), m <= n ->
          proj1_sig e' = proj1_sig e0 ->
          SubValueMC (bevalM m e') (bevalM n e')) ->
        forall (e' : Exp) (n : nat),
          m <= n ->
          proj1_sig e' = e ->
          SubValueMC (bevalM (S m) e') (bevalM (S n) e').
    
    Variable bevalM_continuous_Exp_E : DAlgebra EC_ExpName E E (UP'_SP bevalM_continuous_Exp_P).
    (** Bounded Monadic Evaluation is continuous. **)
    Definition bevalM_continuous : forall m,
      forall (e e' : Exp) (n : nat), m <= n ->
        proj1_sig e' = proj1_sig e ->
        SubValueMC (bevalM m e') (bevalM n e').
    Proof.
      induction m; simpl; intros.
      unfold bevalM; simpl.
      apply inject_i; constructor.
      destruct n.
      elimtype False; inversion H.
      assert (m <= n) by auto with arith.
      destruct e as [e e_UP'].
      apply (projT2 (IndD (P := UP'_SP bevalM_continuous_Exp_P) e e_UP')); auto.
    Qed.

    Definition SVM_Bind_P (i : SubValueM_i) :=
      forall k k', 
        (forall v, SubValueMC (k v) (k' v)) -> 
        SubValueMC ((svm_a i) >>= k) ((svm_b i) >>= k').
    
    Inductive SVM_Bind_Name := svm_bind_name.
    
    Global Instance SVM_Bind_base : iPAlgebra SVM_Bind_Name SVM_Bind_P SubValueM_base.
    Proof.
      econstructor; unfold iAlgebra; intros; apply ind_alg_SVM_base; 
        eauto; unfold SVM_Bind_P; simpl; intros.
      (* SVM_Fail *)
      rewrite bind_fail; apply inject_i; constructor 1.
      (* SVM_Return *)
      repeat rewrite <- left_unit; eauto.
    Qed.
      
    Context {SVM_Bind_SVM: iPAlgebra SVM_Bind_Name SVM_Bind_P SVM}.
    
    Lemma SVM_Bind : forall me me' k k',
      SubValueMC me me' ->
      (forall v, SubValueMC (k v) (k' v)) -> 
      SubValueMC (me >>= k) (me' >>= k').
    Proof.
      intros; apply (ifold_ SVM _ (ip_algebra (iPAlgebra := SVM_Bind_SVM)) _ H); eauto.
    Qed.

    Context {Fun_MT : Functor MT}.
    Context {Mon_MT : Monad MT}.
    Context {Fail_MT : FailMonad MT}.
    
    Section WFValue_Sec.

  (* ============================================== *)
  (* WELL-FORMED VALUES RELATION                     *)
  (* ============================================== *)

      Variable TypContext : Set.

      Record WFValue_i : Set := mk_WFValue_i
        { wfv_S : TypContext;
          wfv_v : Value;
          wfv_T : DType}.

    (** SuperFunctor for Well-Formed Value Relation. **)

      Variable WFV : (WFValue_i -> Prop) -> WFValue_i -> Prop.
      Definition WFValue := iFix WFV.
      Definition WFValueC Sigma V T:= WFValue (mk_WFValue_i Sigma V T).
      Variable funWFV : iFunctor WFV.

      Definition WF_Environment Sigma env env' :=
        P2_Env (WFValueC Sigma) env env'.
      (*fun v T =>
        match T with
          | Some T => WFValueC v T
          | None => False
        end* env env'. *)

    (* Projection doesn't affect WFValue Relation.*)

      Definition WFV_proj1_a_P (i :WFValue_i) :=
        forall a' H, a' = proj1_sig (wfv_v i) ->
          WFValueC (wfv_S i) (exist _ a' H) (wfv_T i).

      Inductive WFV_proj1_a_Name := wfv_proj1_a_name.
      Context {WFV_proj1_a_WFV :
        iPAlgebra WFV_proj1_a_Name WFV_proj1_a_P WFV}.

      Definition WFV_proj1_a :=
        ifold_ WFV _ (ip_algebra (iPAlgebra := WFV_proj1_a_WFV)).

      Definition WFV_proj1_b_P (i :WFValue_i) :=
        forall b' H, b' = proj1_sig (wfv_T i) ->
          WFValueC (wfv_S i) (wfv_v i) (exist _ b' H).

      Inductive WFV_proj1_b_Name := wfv_proj1_b_name.
      Context {WFV_proj1_b_WFV :
        iPAlgebra WFV_proj1_b_Name WFV_proj1_b_P WFV}.

      Definition WFV_proj1_b :=
        ifold_ WFV _ (ip_algebra (iPAlgebra := WFV_proj1_b_WFV)).

      Inductive eval_Soundness_alg_Name := eval_soundness_algname.

      Class ConsExtensionC (TypContext : Set) : Type := 
        { ConsExtension : TypContext -> TypContext -> Prop;
          ConsExtension_id : forall Sigma : TypContext, ConsExtension Sigma Sigma;
          ConsExtension_trans : forall (Sigma Sigma' Sigma'' : TypContext), 
            ConsExtension Sigma' Sigma -> 
            ConsExtension Sigma'' Sigma' -> 
            ConsExtension Sigma'' Sigma
        }.

      Context {TypContextCE : ConsExtensionC TypContext}.

      Definition WFValue_Weaken_P (i : WFValue_i) :=
        forall Sigma', ConsExtension Sigma' (wfv_S i) -> 
          WFValueC Sigma' (wfv_v i) (wfv_T i).

      Inductive WFValue_Weaken_Name := wfvalue_weaken_name.

      Context {WFV_Weaken_WFV : iPAlgebra WFValue_Weaken_Name WFValue_Weaken_P WFV}.

      Definition WFV_Weaken := ifold_ WFV _ (ip_algebra (iPAlgebra := WFV_Weaken_WFV)).

  (* ============================================== *)
  (* WELL-FORMED MONADIC VALUES RELATION            *)
  (* ============================================== *)

      Record WFValueM_i : Set := mk_WFValueM_i
        { wfvm_S : TypContext;
          wfvm_v : ME Value;
          wfvm_T : MT DType}.

    (** SuperFunctor for Well-Formed Value Relation. **)

      Variable WFVM : (WFValueM_i -> Prop) -> WFValueM_i -> Prop.
      Definition WFValueM := iFix WFVM.
      Definition WFValueMC Sigma v T:= WFValueM (mk_WFValueM_i Sigma v T).
      Variable funWFVM : iFunctor WFVM.

      Inductive WFValueM_base (A : WFValueM_i -> Prop) : WFValueM_i -> Prop :=
      | WFVM_Return : forall Sigma v (T : DType) (mt : MT DType),
        WFValueC Sigma v T ->
        {mt' : MT {T' : DType | proj1_sig T = proj1_sig T'} &
          fmap (@proj1_sig _ _) mt' = mt} ->
        WFValueM_base A (mk_WFValueM_i Sigma (return_ v) mt)
      | WFVM_Untyped : forall B Sigma me (mt : MT B),
        WFValueM_base A (mk_WFValueM_i Sigma me (mt >> fail)).

      Definition ind_alg_WFVM_base (P : WFValueM_i -> Prop)
        (H_Return : forall Sigma v T mt,
          WFValueC Sigma v T ->
          {mt' : MT {T' : DType | proj1_sig T = proj1_sig T'} &
            fmap (@proj1_sig _ _) mt' = mt} ->
          P (mk_WFValueM_i Sigma (return_ v) mt))
        (H_Untyped : forall B Sigma me mt,
          P (mk_WFValueM_i Sigma me (mt >> fail)))
        (i : WFValueM_i)
        (e : WFValueM_base P i)
        :
        P i :=
        match e in WFValueM_base _ i return P i with
          | WFVM_Return Sigma v T mt WF_v_T WF_mt => H_Return Sigma v T mt WF_v_T WF_mt
          | WFVM_Untyped B Sigma me mt => H_Untyped B Sigma me mt
        end.

      Definition WFVM_base_ifmap
        (A B : WFValueM_i -> Prop)
        (i : WFValueM_i)
        (f : forall i, A i -> B i)
        (wfvm_v : WFValueM_base A i)
        :
        WFValueM_base B i :=
        match wfvm_v in WFValueM_base _ i return WFValueM_base B i with
          | WFVM_Return Sigma v T mt WF_v_T WF_mt => WFVM_Return B Sigma v T mt WF_v_T WF_mt
          | WFVM_Untyped C Sigma me mt => WFVM_Untyped B C Sigma me mt
        end.

      Global Instance iFun_WFVM_base : iFunctor WFValueM_base.
      constructor 1 with (ifmap := WFVM_base_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; auto.
    Defined.

    Lemma WFVM_Return_T : forall A Sigma v T,
      WFValueC Sigma v T ->
      WFValueM_base A (mk_WFValueM_i Sigma (return_ v) (return_ T)).
    Proof.
      intros; constructor 1 with (T := T); auto.
      exists (return_ (exist (fun T' => proj1_sig T = proj1_sig T')
        T (refl_equal (proj1_sig T)))).
      rewrite fmap_m, <- left_unit; reflexivity.
    Qed.

    Lemma WFVM_Untyped_T : forall A Sigma v,
      WFValueM_base A (mk_WFValueM_i Sigma (return_ v) fail).
    Proof.
      simpl; intros.
      rewrite left_unit with (a := tt) (f := fun _ => fail).
      intros; constructor 2.
    Qed.

    Lemma WFVM_Untyped'_T : forall A (B : Set) (P : B -> Prop)
      Sigma mv (mt : MT {b : B | P b}) (k : B -> MT DType),
      (forall b : B, P b -> k b = fail) ->
      WFValueM_base A (mk_WFValueM_i Sigma mv (fmap (@proj1_sig _ _) mt >>= k)).
    Proof.
      simpl; intros.
      rewrite fmap_m, <- associativity, rewrite_do with (m' := fun x => fail).
      constructor.
      apply functional_extensionality; intros d; destruct d;
        rewrite <- left_unit; auto.
    Qed.

    Lemma WFVM_Untyped''_T : forall A B Sigma mv (mt : MT B)
      (k : B -> MT DType), (forall T : B, k T = fail) ->
      WFValueM_base A (mk_WFValueM_i Sigma mv (mt >>= k)).
    Proof.
      simpl; intros.
      rewrite Eta with (f := k).
      rewrite rewrite_do with (m' := fun x => fail).
      constructor.
      apply functional_extensionality; intros d; auto.
    Qed.

    Lemma WFVM_Untyped3_T : forall A (B C : Set) (P : B -> Prop)
      Sigma mv (mt : MT {b : B | P b}) (k : B -> MT DType)
      (mt' : MT C),
      (forall b : B, P b -> k b = mt' >> fail) ->
      WFValueM_base A (mk_WFValueM_i Sigma mv (fmap (@proj1_sig _ _) mt >>= k)).
    Proof.
      simpl; intros.
      rewrite fmap_m, <- associativity, rewrite_do with (m' := fun x => mt' >> fail).
      unfold wbind; rewrite associativity; constructor.
      apply functional_extensionality; intros d; destruct d;
        rewrite <- left_unit; auto.
    Qed.

    Lemma WFVM_Untyped4_T : forall A (B C : Set) Sigma mv (mt : MT B)
      (k : B -> MT DType) (mt' : MT C), (forall T : B, k T = mt' >> fail) ->
      WFValueM_base A (mk_WFValueM_i Sigma mv (mt >>= k)).
    Proof.
      simpl; intros.
      rewrite Eta with (f := k).
      rewrite rewrite_do with (m' := fun x => mt' >> fail).
      unfold wbind; rewrite associativity; constructor.
      apply functional_extensionality; intros d; auto.
    Qed.

    Definition eval_soundness_P (e : Fix E) (e_UP' : Universal_Property'_fold e) :=
      forall Sigma, WFValueMC Sigma (evalM e) (typeof e).

    Inductive ES_ExpName := es_expname.

    Definition eval_soundness'_P 
      (P_bind : Set)
      (E' : Set -> Set)
      (Fun_E' : Functor E')
      (P : UP'_F E -> UP'_F E' -> P_bind -> TypContext -> Prop)
      (pb : P_bind)
      (typeof_rec : UP'_F E' -> typeofR)
      (eval_rec : Exp -> evalMR)
      (typeof_F : Mixin (UP'_F E') E' typeofR)
      (eval_F : Mixin Exp E evalMR)
      (e : (Fix E') * (Fix E)) 
      (e_UP' : Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e)) := 
      forall 
        (eval_rec_proj : forall e, eval_rec e = eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        (typeof_rec_proj : forall e, typeof_rec e = typeof_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        Sigma (WF_Gamma : P (exist _ _ (proj2 e_UP')) (exist _ _ (proj1 e_UP')) pb Sigma)
          (IHa : forall pb Sigma (a : (UP'_F E' * Exp)) 
            (WF_Gamma : P (snd a) (fst a) pb Sigma), 
            (WFValueMC Sigma  (eval_F eval_rec (out_t_UP' _ _ (proj1_sig (snd a)))) 
              (typeof_F typeof_rec (out_t_UP' _ _ (proj1_sig (fst a)))) -> 
              WFValueMC Sigma (eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig (snd a))))) 
              (typeof_rec (fst a)))),
          WFValueMC Sigma (eval_F eval_rec (out_t_UP' _ _ (snd e))) 
          (typeof_F typeof_rec (out_t_UP' _ _ (fst e))).

  End WFValue_Sec.

  Section Pure_Sound_Sec.

    Variable WFV : (WFValue_i unit -> Prop) -> WFValue_i unit -> Prop.
    Variable funWFV : iFunctor WFV.
    
    Variable WFVM : (WFValueM_i unit -> Prop) -> WFValueM_i unit -> Prop.
    Variable funWFVM : iFunctor WFVM.
  
    Context {Reasonable_MT : Reasonable_Monad MT}.
    
    Definition Pure_Sound_P (i : WFValueM_i unit) :=
      forall T, wfvm_T _ i = return_ T ->
        exists v : Value, 
          wfvm_v _ i = return_ (M := ME) v /\
          WFValueC _ WFV tt v T.
    
    Inductive Pure_Sound_Name := pure_sound_name.
    
    Context {WFV_proj1_b_WFV :
      iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P unit WFV) WFV}.

    Global Instance Pure_Sound_WFVM_base :
      iPAlgebra Pure_Sound_Name Pure_Sound_P (WFValueM_base unit WFV).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV);
        try assumption; unfold Pure_Sound_P; intros.
      (* WFVM_Return' *)
      exists v; split; simpl in *; auto.
      destruct H1 as [mt' mt'_eq]; subst.
      destruct (fmap_exists _ _ _ _ _ H2) as [[T' T_eq] T'_eq].
      simpl in *; subst; auto.
      destruct T0 as [T0 T0_UP'].
      destruct Sigma.
      apply (WFV_proj1_b _ _ _ _ H0); simpl; auto.
      (* WFVM_Untyped' *)
      simpl in H0; apply sym_eq in H0.
      apply FailMonad_Disc in H0; destruct H0; auto.
    Qed.

    Inductive ES'_ExpName := es'_expname.

    Context {WF_MAlg_typeof : WF_MAlgebra Typeof_E}.
    Context {WF_MAlg_eval : WF_MAlgebra evalM_E}.

    Context {eval_soundness_Exp_E : DAlgebra ES_ExpName E E (UP'_SP (eval_soundness_P unit WFVM))}.
    Context {eval_soundness'_Exp_E : forall typeof_rec eval_rec, 
      P2Algebra ES'_ExpName E E E
      (UP'_P2 (eval_soundness'_P unit WFVM unit _ _ (fun _ _ _ _ => True) tt typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := evalM_E _))))}.

    Lemma eval_soundness : forall (e : Exp) Sigma,
      (WFValueMC unit WFVM Sigma (evalM (proj1_sig e)) (typeof (proj1_sig e))).
    Proof.
      intro; rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig e)).
      simpl; unfold typeof, evalM, fold_, mfold, in_t.
      repeat rewrite wf_malgebra; unfold mfold.
      destruct (Ind2 (Ind_Alg := eval_soundness'_Exp_E
        (fun e => typeof (proj1_sig e)) (fun e => evalM (proj1_sig e)))
        _ (proj2_sig e)) as [e' eval_e'].
      unfold eval_soundness'_P in eval_e'.
      intros; eapply eval_e'; intros; auto; try constructor.
      rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto.
      rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto.
      rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig (fst a)) (proj2_sig _)).
      unfold typeof, evalM, mfold; simpl; unfold in_t; 
        repeat rewrite wf_malgebra; unfold mfold; apply H; auto.
    Qed.

    Context {Pure_Sound_WFVM : iPAlgebra Pure_Sound_Name Pure_Sound_P WFVM}.

    Global Instance ConsExtensionUnit : ConsExtensionC unit := 
      { ConsExtension := fun _ _ => True }.
    Proof.
      (* ConsExtension_id *)
      eauto.
      (* ConsExtension_trans *)
      eauto.
    Qed.
        
    Theorem eval_Pure_Sound' :
      forall (e : Exp) (T : DType),
        typeof (proj1_sig e) = return_ T ->
        exists v : Value,
          evalM (proj1_sig e) = return_ (M := ME) v /\
          WFValueC unit WFV tt v T.
    Proof.
      intro; apply (ifold_ WFVM _ (ip_algebra (iPAlgebra := Pure_Sound_WFVM)) _ 
      (eval_soundness e tt)).
    Qed.

  End  Pure_Sound_Sec.

  (* Context {Inj_ME : InjMonad ME}.
    Context {Inj_MT : InjMonad MT}.
    
    Context {State_ME : StateM ME (list Value)}.
    
    Definition State_Sound_P (i : WFValueM_i) :=
      forall T (env : list Value),
        WF_Environment (wfvm_S i) env (wfvm_S i) ->
        wfvm_T i = return_ T ->
        exists v : Value, exists env', exists Sigma',
          (put env) >> wfvm_v i = put env' >> return_ (M := ME) v /\
          WFValueC Sigma' v T.

    Inductive State_Sound_Name := State_Sound_name.

    (* Global Instance State_Sound_WFVM_base :
      iPAlgebra State_Sound_Name State_Sound_P WFValueM_base.
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base;
        try assumption; unfold State_Sound_P; intros.
      exists v; exists env; exists Sigma; split; auto.
      simpl in H2; inversion H2; subst; auto.
      inversion H1.
      (*
      generalize (inj_return (M := MT) T T0 H2);
        intros T_eq; subst; auto.
      simpl in H1. apply sym_eq in H1.
      rewrite left_unit with (a := tt) (f := fun _ => fail) in H1.
      apply FailMonad_Disc in H1; destruct H1; auto.*)
    Qed. *)

    Context {Sub_WFVM_base'_WFVM : Sub_iFunctor WFValueM_base' WFVM}.

    (* Definition WFVM_Abstract_P (i : WFValueM_i) :=
      (exists T : DType,
        exists mt' : MT {T' : DType | proj1_sig T = proj1_sig T'},
          fmap (@proj1_sig _ _) mt' = wfvm_T i) \/
      exists B : Set, exists mt' : MT B, wfvm_T i = mt' >> fail.

    Inductive WFVM_Abstract_Name := WFVM_Abstract_name.

    Global Instance WFVM_Abstract_WFVM_base' :
      iPAlgebra WFVM_Abstract_Name WFVM_Abstract_P WFValueM_base'.
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base';
        try assumption; unfold Pure_Sound_P; intros.
      (* WFVM_Return' *)
      left; exists T.
      destruct H1 as [mt' mt'_eq]; subst.
      exists mt'; reflexivity.
      (* WFVM_Untyped' *)
      right; simpl; repeat eexists _; reflexivity.
    Defined. *)

    Global Instance State_Sound_WFVM_base' :
      iPAlgebra State_Sound_Name State_Sound_P WFValueM_base'.
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base';
        try assumption; unfold State_Sound_P; simpl; intros.
      exists v; exists env; exists Sigma; split; auto.
      destruct H1 as [mt' mt'_eq]; subst.
      destruct (fmap_exists _ _ _ _ _ H3) as [[T' T_eq] T'_eq].
      simpl in *; subst; auto.
      destruct T0 as [T0 T0_UP'].
      apply (WFV_proj1_b _ H0); simpl; auto.
      (* WFVM_Untyped' *)
      simpl in H1; apply sym_eq in H1.
      apply FailMonad_Disc in H1; destruct H1; auto.
    Qed.

    Definition eval_soundness_P (e : Fix E) (e_UP' : Universal_Property'_fold e) :=
      forall Sigma Gamma, WFValueMC Sigma Gamma (evalM e) (typeof e).

    Inductive ES_ExpName := es_expname.

    Definition eval_soundness'_P 
      (P_bind : Set)
      (E' : Set -> Set)
      (Fun_E' : Functor E')
      (P : UP'_F E -> UP'_F E' -> P_bind -> Env DType -> Env DType -> Prop)
      (pb : P_bind)
      (typeof_rec : UP'_F E' -> typeofR)
      (eval_rec : Exp -> evalMR)
      (typeof_F : Mixin (UP'_F E') E' typeofR)
      (eval_F : Mixin Exp E evalMR)
      (e : (Fix E') * (Fix E)) 
      (e_UP' : Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e)) := 
      forall 
        (eval_rec_proj : forall e, eval_rec e = eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        (typeof_rec_proj : forall e, typeof_rec e = typeof_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        Sigma Gamma (WF_Gamma : P (exist _ _ (proj2 e_UP')) (exist _ _ (proj1 e_UP')) pb Sigma Gamma)
          (IHa : forall pb Sigma Gamma (a : (UP'_F E' * Exp)) 
            (WF_Gamma : P (snd a) (fst a) pb Sigma Gamma), 
            (WFValueMC Sigma Gamma  (eval_F eval_rec (out_t_UP' _ _ (proj1_sig (snd a)))) 
              (typeof_F typeof_rec (out_t_UP' _ _ (proj1_sig (fst a)))) -> 
              WFValueMC Sigma Gamma (eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig (snd a))))) 
              (typeof_rec (fst a)))),
          WFValueMC Sigma Gamma (eval_F eval_rec (out_t_UP' _ _ (snd e))) 
          (typeof_F typeof_rec (out_t_UP' _ _ (fst e))).

    Inductive ES'_ExpName := es'_expname.

    Context {eval_soundness_Exp_E : DAlgebra ES_ExpName E E (UP'_SP eval_soundness_P)}.
    Context {eval_soundness'_Exp_E : forall typeof_rec eval_rec, 
      P2Algebra ES'_ExpName E E E
      (UP'_P2 (eval_soundness'_P unit _ _ (fun _ _ _ _ _ => True) tt typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := evalM_E _))))}.

    Context {Pure_Sound_WFVM : iPAlgebra Pure_Sound_Name Pure_Sound_P WFVM}.

    Theorem eval_Pure_Sound :
      forall (e : Exp) (T : DType),
        typeof (proj1_sig e) = return_ T ->
        exists v : Value, exists Sigma : Env DType,
          evalM (proj1_sig e) = return_ (M := ME) v /\
          WFValueC Sigma v T.
    Proof.
      intro e.
      apply (ifold_ WFVM _ (ip_algebra (iPAlgebra := Pure_Sound_WFVM)) _
        (proj2_sig (IndD (P := UP'_P eval_soundness_P) (proj1_sig e) (proj2_sig e)) nil nil)).
    Qed.

    Lemma eval_soundness : forall (e : Exp) Sigma Gamma,
      (WFValueMC Sigma Gamma (evalM (proj1_sig e)) (typeof (proj1_sig e))).
    Proof.
      intro; rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig e)).
      simpl; unfold typeof, evalM, fold_, mfold, in_t.
      repeat rewrite wf_malgebra; unfold mfold.
      destruct (Ind2 (Ind_Alg := eval_soundness'_Exp_E
        (fun e => typeof (proj1_sig e)) (fun e => evalM (proj1_sig e)))
        _ (proj2_sig e)) as [e' eval_e'].
      unfold eval_soundness'_P in eval_e'.
      intros; eapply eval_e'; intros; auto; try constructor.
      rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto.
      rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto.
      rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig (fst a)) (proj2_sig _)).
      unfold typeof, evalM, mfold; simpl; unfold in_t; 
        repeat rewrite wf_malgebra; unfold mfold; apply H; auto.
    Qed.

    Context {State_Sound_WFVM : iPAlgebra State_Sound_Name State_Sound_P WFVM}.

    Theorem eval_State_Sound :
      forall (Sigma : Env DType) (e : Exp) (T : DType) (env : Env Value),
        WF_Environment Sigma env Sigma ->
        typeof (proj1_sig e) = return_ T ->
        exists v : Value,
          exists env' : Env Value,
            exists Sigma' : Env DType,
              (put env) >> evalM (proj1_sig e) = put env' >> return_ (M := ME) v /\
              WFValueC Sigma' v T.
    Proof.
      intros Sigma e.
      apply (ifold_ WFVM _ (ip_algebra (iPAlgebra := State_Sound_WFVM)) _ 
        (eval_soundness e Sigma nil)).
    Qed. *)

    (* Inductive WFValueM_base' (A : WFValueM_i -> Prop) : WFValueM_i -> Prop :=
    | WFVM_Return' : forall v T,
      WFValueC v T ->
      WFValueM_base' A (mk_WFValueM_i (return_ v) (return_ T))
    | WFVM_Untyped' : forall me,
      WFValueM_base' A (mk_WFValueM_i me fail)
    | WFVM_Tind' : forall mv mt kv kt,
      A (mk_WFValueM_i mv mt) ->
      (forall v T, WFValueC v T -> A (mk_WFValueM_i (kv v) (kt T))) ->
      WFValueM_base' A (mk_WFValueM_i (do v <- mv; kv v) (do t <- mt; kt t)).

    Definition ind_alg_WFVM_base' (P : WFValueM_i -> Prop)
      (H_Return' : forall v T,
        WFValueC v T ->
        P (mk_WFValueM_i (return_ v) (return_ T)))
      (H_Untyped' : forall me,
        P (mk_WFValueM_i me fail))
      (H_Bind' : forall mv mt kv kt,
        P (mk_WFValueM_i mv mt) ->
        (forall v T, WFValueC v T -> P (mk_WFValueM_i (kv v) (kt T))) ->
        P (mk_WFValueM_i (do v <- mv; kv v) (do t <- mt; kt t)))
      (i : WFValueM_i)
      (e : WFValueM_base' P i)
      :
      P i :=
      match e in WFValueM_base' _ i return P i with
        | WFVM_Return' v T WF_v_T => H_Return' v T WF_v_T
        | WFVM_Untyped' me => H_Untyped' me
        | WFVM_Tind' mv mt kv kt P_mv_mt P_kv_kt =>
          H_Bind' mv mt kv kt P_mv_mt P_kv_kt
      end.

    Definition WFVM_base'_ifmap
      (A B : WFValueM_i -> Prop)
      (i : WFValueM_i)
      (f : forall i, A i -> B i)
      (wfvm_v : WFValueM_base' A i)
      :
      WFValueM_base' B i :=
      match wfvm_v in WFValueM_base' _ i return WFValueM_base' B i with
        | WFVM_Return' v T WF_v_T => WFVM_Return' B v T WF_v_T
        | WFVM_Untyped' me => WFVM_Untyped' B me
        | WFVM_Tind' mv mt kv kt P_mv_mt P_kv_kt =>
          WFVM_Tind' B mv mt kv kt (f _ P_mv_mt) (fun v T WF_v_T => f _ (P_kv_kt v T WF_v_T))
      end.

    Global Instance iFun_WFVM_base' : iFunctor WFValueM_base'.
      constructor 1 with (ifmap := WFVM_base'_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; auto.
      unfold id; apply f_equal; apply functional_extensionality_dep; intros.
      apply functional_extensionality_dep; intros.
      apply functional_extensionality_dep; intros.
      reflexivity.
    Defined.

    Context {Sub_WFVM_base'_WFVM : Sub_iFunctor WFValueM_base' WFVM}.


    Global Instance Pure_Return_WFVM_base' :
      iPAlgebra Pure_Soundness_Name Pure_Return_P WFValueM_base'.
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base';
        try assumption; unfold Pure_Return_P; intros.
      exists v; split; auto.
      generalize (inj_return (M := MT) (Some T) (Some T0) H1);
        intros T_eq; injection T_eq; intros; subst; auto.
      generalize (inj_return (M := MT) _ _ H0); intros;
        discriminate.
      simpl in H2.

    Qed.*)


End Names.

    (* Lemma bevalM_continuous' : forall m e' n vm vn,  *)
    (*   m <= n ->  *)
    (*   bevalM m e' = return_ vm ->  *)
    (*   bevalM n e' = return_ vn ->  *)
    (*   SubValueC vm vn. *)
    (*   intros. *)
    (*   destruct (bevalM_continuous m e') as [Mm Mm_eq]. *)
    (*   rewrite <- Mm_eq in H0. *)
    (*   (* We should be able to get *)
    (*      some leverage by using strategy trees here.  *)
    (*      Alternatively, we could show that this holds *)
    (*      for evaluation functions that uses  the Id monad.  *)
    (*      *) *)
    (* Abort. *)


(*   (* ============================================== *) *)
(*   (* WELL-FORMED VALUES RELATION                     *) *)
(*   (* ============================================== *) *)

(*     Record WFValue_i : Set := mk_WFValue_i  *)
(*       {wfv_v : Value; *)
(*         wfv_T : DType}. *)

(*     (** SuperFunctor for Well-Formed Value Relation. **) *)

(*     Variable WFV : (WFValue_i -> Prop) -> WFValue_i -> Prop. *)
(*     Definition WFValue := iFix WFV. *)
(*     Definition WFValueC V T:= WFValue (mk_WFValue_i V T). *)
(*     Variable funWFV : iFunctor WFV. *)








(*     Variable WF_WFV_Tot_WFV : Sub_iFunctor WFValue_Bot WFV. *)

(*     Definition WF_Environment env env' := *)
(*       P2_Env (fun v T =>  *)
(*         match T with  *)
(*           | Some T => WFValueC v T *)
(*           | None => False *)
(*         end) env env'. *)

(*     (* Projection doesn't affect WFValue Relation.*) *)

(*      Definition WFV_proj1_a_P (i :WFValue_i) :=  *)
(*       forall a' H, a' = proj1_sig (wfv_v i) ->   *)
(*         WFValueC (exist _ a' H) (wfv_T i). *)

(*      Inductive WFV_proj1_a_Name := wfv_proj1_a_name. *)
(*      Context {WFV_proj1_a_WFV : *)
(*        iPAlgebra WFV_proj1_a_Name WFV_proj1_a_P WFV}.     *)

(*      Definition WFV_proj1_a :=  *)
(*        ifold_ WFV _ (ip_algebra (iPAlgebra := WFV_proj1_a_WFV)). *)

(*      Global Instance WFV_proj1_a_Bot :  *)
(*        iPAlgebra WFV_proj1_a_Name WFV_proj1_a_P WFValue_Bot. *)
(*        econstructor; intros. *)
(*        unfold iAlgebra; intros; unfold WFV_proj1_a_P. *)
(*        inversion H; subst; simpl; intros. *)
(*        apply (inject_i (subGF := WF_WFV_Tot_WFV)); constructor; simpl; congruence. *)
(*      Defined. *)

(*      Definition WFV_proj1_b_P (i :WFValue_i) :=  *)
(*       forall b' H, b' = proj1_sig (wfv_T i) ->   *)
(*         WFValueC (wfv_v i) (exist _ b' H). *)

(*      Inductive WFV_proj1_b_Name := wfv_proj1_b_name. *)
(*      Context {WFV_proj1_b_WFV : *)
(*        iPAlgebra WFV_proj1_b_Name WFV_proj1_b_P WFV}.     *)

(*      Definition WFV_proj1_b :=  *)
(*        ifold_ WFV _ (ip_algebra (iPAlgebra := WFV_proj1_b_WFV)). *)

(*      Global Instance WFV_proj1_b_Bot :  *)
(*        iPAlgebra WFV_proj1_b_Name WFV_proj1_b_P WFValue_Bot. *)
(*        econstructor; intros. *)
(*        unfold iAlgebra; intros; unfold WFV_proj1_b_P. *)
(*        inversion H; subst; simpl; intros. *)
(*        apply (inject_i (subGF := WF_WFV_Tot_WFV)); constructor; simpl; congruence. *)
(*      Defined. *)

(*   (* ============================================== *) *)
(*   (* Evaluation preserves Well-Formedness           *) *)
(*   (* ============================================== *) *)

(*     Definition WF_Value_continuous_P i :=  *)
(*       forall T, WFValueC (sv_b i) T -> WFValueC (sv_a i) T. *)

(*     Inductive WFV_ContinuousName : Set := wfv_continuousname. *)

(*     Context {WF_Value_continous_alg : iPAlgebra WFV_ContinuousName WF_Value_continuous_P SV}. *)

(*     Global Instance WFV_Value_continuous_refl  :  *)
(*       iPAlgebra WFV_ContinuousName WF_Value_continuous_P SubValue_refl. *)
(*       constructor; unfold iAlgebra; intros. *)
(*       inversion H; subst. *)
(*       unfold WF_Value_continuous_P; simpl; intros. *)
(*       unfold WFValueC in H1. *)
(*       destruct v; apply (WFV_proj1_a _ H1); eauto. *)
(*     Qed. *)

(*     Global Instance WFV_Value_continuous_Bot  :  *)
(*       iPAlgebra WFV_ContinuousName WF_Value_continuous_P SubValue_Bot. *)
(*       constructor; unfold iAlgebra; intros. *)
(*       inversion H; subst. *)
(*       unfold WF_Value_continuous_P; simpl; intros. *)
(*       apply inject_i; constructor; auto. *)
(*     Qed. *)

(*     Lemma WF_Value_continous : forall v v', *)
(*       SubValueC v v' ->  *)
(*       WF_Value_continuous_P (mk_SubValue_i v v'). *)
(*       intros; apply (ifold_ SV); try assumption. *)
(*       apply ip_algebra. *)
(*     Qed. *)

(*     Lemma WF_Value_beval : forall m n, *)
(*       forall (e : Exp) gamma gamma' T *)
(*         (Sub_G_G' : Sub_Environment gamma' gamma), *)
(*         m <= n ->  *)
(*         WFValueC (beval n e gamma) T ->  *)
(*         WFValueC (beval m e gamma') T. *)
(*       intros; eapply WF_Value_continous. *)
(*       eapply beval_continuous; try eassumption. *)
(*       eassumption. *)
(*     Qed. *)

(*     Variable (WF_MAlg_typeof : WF_MAlgebra Typeof_E). *)
(*     Variable (WF_MAlg_eval : WF_MAlgebra eval_E). *)

(*     Definition eval_alg_Soundness_P  *)
(*       (P_bind : Set) *)
(*       (P : P_bind -> Env Value -> Prop) *)
(*       (E' : Set -> Set) *)
(*       (Fun_E' : Functor E') *)
(*       (pb : P_bind) *)
(*       (typeof_rec : UP'_F E' -> typeofR) *)
(*       (eval_rec : Exp -> evalR) *)
(*       (typeof_F : Mixin (UP'_F E') E' typeofR) *)
(*       (eval_F : Mixin Exp E evalR) *)
(*       (e : (Fix E') * (Fix E))  *)
(*       (e_UP' : Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e)) :=  *)
(*       forall  *)
(*         (eval_rec_proj : forall e, eval_rec e = eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e)))) *)
(*         (typeof_rec_proj : forall e, typeof_rec e = typeof_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e)))) *)
(*       gamma'' (WF_gamma'' : P pb gamma'') *)
(*       (IHa : forall pb gamma'' (WF_gamma'' : P pb gamma'') (a : (UP'_F E' * Exp)),  *)
(*           (forall T,  *)
(*             typeof_F typeof_rec (out_t_UP' _ _ (proj1_sig (fst a))) = Some T -> *)
(*             WFValueC (eval_F eval_rec (out_t_UP' _ _ (proj1_sig (snd a))) gamma'') T) ->  *)
(*           forall T, typeof_rec (fst a) = Some T ->  *)
(*             WFValueC (eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig (snd a)))) gamma'') T), *)
(*         forall T : DType, *)
(*           typeof_F typeof_rec (out_t_UP' _ _ (fst e)) = Some T -> *)
(*           WFValueC (eval_F eval_rec (out_t_UP' _ _ (snd e)) gamma'') T. *)

(*     Inductive eval_Soundness_alg_Name := eval_soundness_algname. *)

(*     Variable eval_Soundness_alg_F : forall typeof_rec eval_rec, *)
(*       P2Algebra eval_Soundness_alg_Name E E E (UP'_P2 (eval_alg_Soundness_P unit  *)
(*         (fun _ _ => True) _ _ tt *)
(*         typeof_rec eval_rec *)
(*         (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := eval_E _)))). *)

(*     Definition eval_soundness_P (e : Fix E) (e_UP' : Universal_Property'_fold e) := *)
(*       forall gamma'' T, typeof e = Some T ->  *)
(*         WFValueC (eval e gamma'') T. *)

(*     Lemma eval_Soundness : forall (e : Exp), *)
(*       forall gamma'' T, typeof (proj1_sig e) = Some T ->  *)
(*         WFValueC (eval (proj1_sig e) gamma'') T. *)
(*       intros.       *)
(*       rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig e)). *)
(*       simpl; unfold typeof, eval, fold_, mfold, in_t. *)
(*       rewrite wf_malgebra; unfold mfold. *)
(*       destruct (Ind2 (Ind_Alg := eval_Soundness_alg_F  *)
(*         (fun e => typeof (proj1_sig e)) (fun e => eval (proj1_sig e))) *)
(*         _ (proj2_sig e)) as [e' eval_e']. *)
(*       unfold eval_alg_Soundness_P in eval_e'. *)
(*       eapply eval_e'; intros; auto; try constructor. *)
(*       rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto. *)
(*       rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto. *)
(*       unfold eval, mfold; simpl; unfold in_t; rewrite wf_malgebra; unfold mfold; apply H0; auto. *)
(*       rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig (fst a)) (proj2_sig (fst a))) in H1. *)
(*       simpl in H1; unfold typeof, mfold, in_t in H1. *)
(*       rewrite wf_malgebra in H1; apply H1. *)
(*       rewrite <- (@in_out_inverse _ _ (proj1_sig e) (proj2_sig e)) in H. *)
(*       simpl in H; unfold typeof, mfold, in_t in H; simpl in H. *)
(*       rewrite <- wf_malgebra. *)
(*       simpl; unfold out_t_UP'. *)
(*       rewrite Fusion with (g := (fmap in_t)).  *)
(*       apply H. *)
(*       exact (proj2_sig _). *)
(*       intros; repeat rewrite fmap_fusion; reflexivity. *)
(*     Qed. *)

(* End Names. *)

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
